/*************************************************************************
* (C) 2008 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: parse.c,v 1.14 2014/01/30 08:30:47 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2014
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


/*
 * cfgReadConfig() - Read file and add new item at config root
 *
 * @f = File resource
 * @cfg = Config root
 * return: -1 error or 0 ok
 */
int
cfgReadConfig(FILE *f, cfg_root_t * __restrict cfg)
{
	char line[BUFSIZ];
	struct tagCfg *av = NULL;
	int flg = 0;
	char *psAttr, *psVal, szSection[STRSIZ] = { 0 };

	if (!f || !cfg) {
		cfg_SetErr(EINVAL, "Invalid parameter(s)");
		return -1;
	}

	while (!feof(f)) {
		memset(line, 0, sizeof line);
		fgets(line, sizeof line - 1, f);
#ifdef SUPPORT_USER_EOF
		/* check for user end-of-file */
		if (line[0] == '.' && line[1] == '\n')
			break;
#endif
		if (!(psAttr = strpbrk(line, "\r\n"))) {
			/* skip line, too long */
			continue;
		} else {
			*psAttr = 0;
			str_Trim(line);
		}

		if (flg) {
			/* continues line */
			if (!av)
				continue;
			else
				psAttr = line + strlen(line) - 1;
			if (*psAttr == '\\')
				*psAttr = 0;
			else
				flg = 0;
			/* concat line to value */
			AIT_SET_STRCAT(&av->cfg_val, line);
			if (!flg && AIT_ADDR(&av->cfg_val))
				str_Unquot((char*) AIT_GET_STR(&av->cfg_val));
			continue;
		}

		/* *NEW PAIR* alloc new pair element */
		av = e_malloc(sizeof(struct tagCfg));
		if (!av) {
			cfg_SetErr(elwix_GetErrno(), "%s", elwix_GetError());
			return -1;
		} else {
			memset(av, 0, sizeof(struct tagCfg));
			CFG_RC_LOCK(cfg);
			TAILQ_INSERT_TAIL(cfg, av, cfg_next);
			CFG_RC_UNLOCK(cfg);
		}

		/* check for continues line */
		psAttr = line + (*line ? strlen(line) : 1) - 1;
		if (*psAttr == '\\') {
			*psAttr = 0;
			flg = 1;
		}

		/* check for comment or empty line */
		if (!*line || *line == '#' || *line == ';') {
			AIT_SET_STR(&av->cfg_val, line);
			continue;
		}
		/* section */
		if (*line == '[') {
			psAttr = line + strlen(line) - 1;
			if (*psAttr == ']') {
				*psAttr = 0; 
				flg = 0;
				strlcpy(szSection, line + 1, sizeof szSection);
				AIT_SET_STR(&av->cfg_sec, line);
			} else
				EDEBUG(7, "Ignore section '%s' ... not found ']'", line);
			continue;
		}
		/* parse pair */
		if (!(psAttr = strchr(line, '='))) {
			AIT_SET_STR(&av->cfg_val, line);
			EDEBUG(7, "Ignore a/v '%s' ... not found '='", line);
			continue;
		} else {
			*psAttr = 0;
			psVal = psAttr + 1;
			psAttr = line;
		}

		/* if exists, added section name to element */
		if (*szSection) {
			AIT_SET_STR(&av->cfg_sec, szSection);
			AIT_KEY(&av->cfg_sec) = crcFletcher16(AIT_GET_LIKE(&av->cfg_sec, u_short*), 
					E_ALIGN(AIT_LEN(&av->cfg_sec) - 1, 2) / 2);
		}

		str_RTrim(psAttr);
		str_LTrim(psVal);
		if (!flg)
			str_Unquot(psVal);
		AIT_SET_STR(&av->cfg_val, psVal);
		AIT_SET_STR(&av->cfg_attr, psAttr);
		AIT_KEY(&av->cfg_attr) = crcFletcher16(AIT_GET_LIKE(&av->cfg_attr, u_short*), 
				E_ALIGN(AIT_LEN(&av->cfg_attr) - 1, 2) / 2);

		CFG_RC_LOCK(cfg);
		RB_INSERT(tagRC, cfg, av);
		CFG_RC_UNLOCK(cfg);
	}

	return 0;
}

/*
 * cfgWriteConfig() - Write config from memory
 *
 * @f = File handle
 * @cfg = Config root
 * @whitespace = Additional whitespace characters to file
 * return: -1 error or 0 ok
 */
int
cfgWriteConfig(FILE *f, cfg_root_t * __restrict cfg, int whitespace)
{
	struct tagCfg *av;
	time_t tim;
	char line[BUFSIZ] = { 0 }, szSection[STRSIZ] = { [0 ... STRSIZ - 1] = 0 };

	if (!f || !cfg) {
		cfg_SetErr(EINVAL, "Invalid parameter(s)");
		return -1;
	}

	if (whitespace) {
		time(&tim);
		memset(line, 0, sizeof line);
		strftime(line, sizeof line, "(UTC) %Y-%m-%d %H:%M:%S", gmtime(&tim));
		cfg_Write(f, "## Config auto-generated at :: %s ##\n", line);
	}

	CFG_RC_LOCK(cfg);
	RB_FOREACH(av, tagRC, cfg) {
		/* empty lines or comment */
		if (AIT_ISEMPTY(&av->cfg_attr)) {
			if (AIT_ISEMPTY(&av->cfg_val))
				continue;
			strlcpy(line, AIT_GET_STR(&av->cfg_val), sizeof line);
			goto skip_sec;
		}

		/* section [] */
		if (!AIT_ISEMPTY(&av->cfg_sec) && AIT_ADDR(&av->cfg_sec) && 
				strcmp(AIT_GET_STRZ(&av->cfg_sec), szSection)) {
			strlcpy(szSection, AIT_GET_STR(&av->cfg_sec), sizeof szSection);
			if (!cfg_Write(f, "\n[%s]\n", AIT_GET_STR(&av->cfg_sec))) {
				LOGERR;
				CFG_RC_UNLOCK(cfg);
				return -1;
			}
		} else if (AIT_ISEMPTY(&av->cfg_sec) && *szSection) {
			memset(szSection, 0, sizeof szSection);
			if (!cfg_Write(f, "\n[]\n")) {
				LOGERR;
				CFG_RC_UNLOCK(cfg);
				return -1;
			}
		}

		/* build line */
		memset(line, 0, sizeof line);
		if (!AIT_ISEMPTY(&av->cfg_attr) && AIT_TYPE(&av->cfg_attr) == string) {
			strlcpy(line, AIT_GET_STRZ(&av->cfg_attr), sizeof line);
			if (whitespace)
				strlcat(line, " = ", sizeof line);
			else
				strlcat(line, "=", sizeof line);
		}
		if (!AIT_ISEMPTY(&av->cfg_val) && AIT_TYPE(&av->cfg_val) == string)
			strlcat(line, AIT_GET_STRZ(&av->cfg_val), sizeof line);
skip_sec:
		/* write */
		if (!cfg_Write(f, "%s\n", line)) {
			LOGERR;
			CFG_RC_UNLOCK(cfg);
			return -1;
		}
	}
	CFG_RC_UNLOCK(cfg);

	if (whitespace) {
		time(&tim);
		memset(line, 0, sizeof line);
		strftime(line, sizeof line, "(UTC) %Y-%m-%d %H:%M:%S", gmtime(&tim));
		cfg_Write(f, "\n## Config was saved at :: %s ##\n", line);
	}

	return 0;
}

/*
 * cfgConcatConfig() - Concat two configs into one
 *
 * @cfg = Config root
 * @add_cfg = Concated config will be destroy after merge
 * return: -1 error or 0 ok
 */
int
cfgConcatConfig(cfg_root_t * __restrict cfg, cfg_root_t * __restrict add_cfg)
{
	struct tagCfg *item;

	if (!cfg || !add_cfg)
		return -1;

	CFG_RC_LOCK(add_cfg);
	CFG_RC_LOCK(cfg);

	/* concat lists & red-black trees */
	TAILQ_FOREACH(item, add_cfg, cfg_next) {
		TAILQ_INSERT_TAIL(cfg, item, cfg_next);
		RB_INSERT(tagRC, cfg, item);
	}

	CFG_RC_UNLOCK(cfg);

	TAILQ_INIT(add_cfg);
	RB_INIT(add_cfg);
	CFG_RC_UNLOCK(add_cfg);
	pthread_mutex_destroy(&add_cfg->rc_mtx);
	return 0;
}

/*
 * cfgMergeConfig() - Marge two list in one cfg and destroy add_cfg
 *
 * @cfg = Config root of main list
 * @add_cfg = Merged config will be destroy after merge
 * return: -1 error or 0 ok
 */
int
cfgMergeConfig(cfg_root_t * __restrict cfg, cfg_root_t * __restrict add_cfg)
{
	struct tagCfg *item, *merge, *add_next, *next;
	int flg;

	if (!cfg || !add_cfg)
		return -1;

	CFG_RC_LOCK(add_cfg);
	CFG_RC_LOCK(cfg);

	/* merge lists */
	TAILQ_FOREACH_SAFE(item, add_cfg, cfg_next, add_next) {
		flg = 0;
		TAILQ_FOREACH_SAFE(merge, cfg, cfg_next, next) {
			if (AIT_ISEMPTY(&merge->cfg_sec) && AIT_ISEMPTY(&item->cfg_sec)) {
				flg = 1;
				break;
			}
			if (!AIT_ISEMPTY(&merge->cfg_sec) && !AIT_ISEMPTY(&item->cfg_sec) && 
					AIT_ADDR(&merge->cfg_sec) && AIT_ADDR(&item->cfg_sec) &&
					!strcmp(AIT_GET_STR(&merge->cfg_sec), AIT_GET_STR(&item->cfg_sec))) {
				flg = 1;
				break;
			}
		}

		if (!flg)
			TAILQ_INSERT_TAIL(cfg, item, cfg_next);
		else
			TAILQ_INSERT_AFTER(cfg, merge, item, cfg_next);
		RB_INSERT(tagRC, cfg, item);
	}

	CFG_RC_UNLOCK(cfg);

	TAILQ_INIT(add_cfg);
	RB_INIT(add_cfg);
	CFG_RC_UNLOCK(add_cfg);
	pthread_mutex_destroy(&add_cfg->rc_mtx);
	return 0;
}

/*
 * cfgReadLines() - Read custom lines and add new item at config root
 *
 * @f = File resource
 * @delim = Custom delimiter, if =NULL default is '='
 * @end = Custom user end of file, if =NULL default is EOF
 * @cfg = Config root
 * return: -1 error or 0 ok
 */
int
cfgReadLines(FILE *f, const char *delim, const char *end, cfg_root_t * __restrict cfg)
{
	char line[BUFSIZ];
	struct tagCfg *d, *av = NULL;
	char *p, *psSec, *psAttr, *psVal;

	if (!cfg)
		return -1;
	if (!delim)
		delim = ATR_LINES_DELIM;

	while (!feof(f)) {
		psSec = psAttr = psVal = NULL;
		memset(line, 0, sizeof line);
		fgets(line, sizeof line - 1, f);
		/* check for user end-of-file */
		if (strspn(line, end))
			break;

		if (!(psAttr = strpbrk(line, "\r\n"))) {
			/* skip line, too long */
			continue;
		} else {
			*psAttr = 0;
			str_Trim(line);
			if (!*line)
				continue;
		}

		if (!av_MakeExt(line, delim, &p, &psVal))
			continue;
		else {
			str_RTrim(p);
			str_LTrim(psVal);
		}
		if (!av_MakeExt(p, SEC_LINES_DELIM, &psSec, &psAttr))
			psAttr = p;

		/* *NEW PAIR* alloc new pair element */
		av = e_malloc(sizeof(struct tagCfg));
		if (!av) {
			LOGERR;
			return -1;
		} else
			memset(av, 0, sizeof(struct tagCfg));

		if (psSec) {
			AIT_SET_STR(&av->cfg_sec, psSec);
			AIT_KEY(&av->cfg_sec) = crcFletcher16(AIT_GET_LIKE(&av->cfg_sec, u_short*), 
					E_ALIGN(AIT_LEN(&av->cfg_sec) - 1, 2) / 2);
		}
		if (psVal)
			AIT_SET_STR(&av->cfg_val, psVal);
		AIT_SET_STR(&av->cfg_attr, psAttr);
		AIT_KEY(&av->cfg_attr) = crcFletcher16(AIT_GET_LIKE(&av->cfg_attr, u_short*), 
				E_ALIGN(AIT_LEN(&av->cfg_attr) - 1, 2) / 2);

		CFG_RC_LOCK(cfg);
		/* find & delete duplicates */
		if ((d = RB_FIND(tagRC, cfg, av))) {
			RB_REMOVE(tagRC, cfg, d);
			TAILQ_REMOVE(cfg, d, cfg_next);

			AIT_FREE_VAL(&d->cfg_val);
			AIT_FREE_VAL(&d->cfg_attr);
			AIT_FREE_VAL(&d->cfg_sec);
			e_free(d);
		}

		TAILQ_INSERT_TAIL(cfg, av, cfg_next);
		RB_INSERT(tagRC, cfg, av);
		CFG_RC_UNLOCK(cfg);
	}

	return 0;
}

/*
 * cfgWriteLines() - Write custom lines and export data to variable
 *
 * @f = File resource
 * @delim = Custom delimiter, if =NULL default is '='
 * @eol = End of line string, if =NULL default is "\n"
 * @section = Export only section, if =NULL default is all
 * @cfg = Config root
 * return: =NULL error or !=NULL exported data, must be free after use with ait_freeVar()
 */
ait_val_t *
cfgWriteLines(FILE *f, const char *delim, const char *eol, const char *section, cfg_root_t * __restrict cfg)
{
	ait_val_t *v = NULL;
	struct tagCfg *av;

	if (!cfg)
		return NULL;
	if (!delim)
		delim = ATR_LINES_DELIM;
	if (!eol)
		eol = EOL_LINES_DELIM;
	if (!(v = ait_allocVar())) {
		cfg_SetErr(elwix_GetErrno(), "%s", elwix_GetError());
		return NULL;
	} else
		AIT_INIT_VAL2(v, string);

	TAILQ_FOREACH(av, cfg, cfg_next) {
		if (AIT_ISEMPTY(&av->cfg_attr))
			continue;
		if (section) {
			if (!AIT_ISEMPTY(&av->cfg_sec) && *section)
				continue;
			if (strcmp(section, AIT_GET_STRZ(&av->cfg_sec)))
				continue;
		}

		if (!AIT_ISEMPTY(&av->cfg_sec)) {
			AIT_SET_STRCAT(v, AIT_GET_STR(&av->cfg_sec));
			AIT_SET_STRCAT(v, SEC_LINES_DELIM);
		}
		AIT_SET_STRCAT(v, AIT_GET_STR(&av->cfg_attr));
		AIT_SET_STRCAT(v, delim);
		if (!AIT_ISEMPTY(&av->cfg_val))
			AIT_SET_STRCAT(v, AIT_GET_STR(&av->cfg_val));
		AIT_SET_STRCAT(v, eol);
	}

	if (f)
		fputs(AIT_GET_STR(v), f);
	return v;
}
