/*************************************************************************
* (C) 2008 AITNET ltd - Sofia/Bulgaria - <misho@aitbg.com>
*  by Michael Pounov <misho@openbsd-bg.org>
*
* $Author: misho $
* $Id: queue.c,v 1.16 2014/04/14 19:58:21 misho Exp $
*
**************************************************************************
The ELWIX and AITNET software is distributed under the following
terms:

All of the documentation and software included in the ELWIX and AITNET
Releases is copyrighted by ELWIX - Sofia/Bulgaria <info@elwix.org>

Copyright 2004 - 2014
	by Michael Pounov <misho@elwix.org>.  All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. All advertising materials mentioning features or use of this software
   must display the following acknowledgement:
This product includes software developed by Michael Pounov <misho@elwix.org>
ELWIX - Embedded LightWeight unIX and its contributors.
4. Neither the name of AITNET nor the names of its contributors
   may be used to endorse or promote products derived from this software
   without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AITNET AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
*/
#include "global.h"


static inline struct tagCfg *
_selectAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	struct tagCfg fav, *c;

	if (!cfg)
		return NULL;
	else
		memset(&fav, 0, sizeof fav);

	if (csSec && *csSec)
		AIT_KEY(&fav.cfg_sec) = crcFletcher16((u_short*) csSec, 
				E_ALIGN(strlen(csSec), 2) / 2);
	if (csAttr)
		AIT_KEY(&fav.cfg_attr) = crcFletcher16((u_short*) csAttr, 
				E_ALIGN(strlen(csAttr), 2) / 2);

	if (!csAttr)
		return RB_NFIND(tagRC, cfg, &fav);
	else {
		c = RB_FIND(tagRC, cfg, &fav);
		if (!c)
			return NULL;	/* not found */
		do {
			if (!strcmp(AIT_GET_STR(&c->cfg_attr), csAttr))
				return c;	/* FOUND! */
		} while ((c = RB_NEXT(tagRC, cfg, c)) && c && !cfg_tree_cmp(c, &fav));
		return NULL;	/* not found */
	}
}

/* --------------------------------------------------------------- */

/*
 * cfg_getSection() - Get entire section attributes into array
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}]
 * return: NULL not found or !=NULL allocated array, must free with array_Destroy() after use!
 */
array_t *
cfg_getSection(cfg_root_t * __restrict cfg, const char *csSec)
{
	array_t *arr = NULL;
	struct tagCfg *av, fav;

	if (!cfg) {
		cfg_SetErr(EINVAL, "Invalid argument(s)");
		return NULL;
	} else
		memset(&fav, 0, sizeof fav);
	if (csSec && !*csSec)
		csSec = NULL;

	if (csSec && *csSec)
		AIT_KEY(&fav.cfg_sec) = crcFletcher16((u_short*) csSec, 
				E_ALIGN(strlen(csSec), 2) / 2);

	av = RB_NFIND(tagRC, cfg, &fav);
	if (!av)
		return NULL;
	if (csSec) {
		if (AIT_ISEMPTY(&av->cfg_sec) || strcmp(AIT_GET_STR(&av->cfg_sec), csSec))
			return NULL;
	} else {
		if (!AIT_ISEMPTY(&av->cfg_sec))
			return NULL;
	}

	arr = array_Init(1);
	if (!arr) {
		cfg_SetErr(elwix_GetErrno(), "%s", elwix_GetError());
		return NULL;
	} else
		array_Push(arr, av, 0);

	while ((av = RB_NEXT(tagRC, cfg, av)) && av) {
		if (csSec) {
			if (AIT_ISEMPTY(&av->cfg_sec) || strcmp(AIT_GET_STR(&av->cfg_sec), csSec))
				break;
		} else {
			if (!AIT_ISEMPTY(&av->cfg_sec))
				break;
		}

		array_Push(arr, av, 0);
	}

	return arr;
}

/*
 * cfg_findAttribute() - Find attribute position in config file
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}]
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: 0 not found item, -1 error or >0 position in list
 */
int
cfg_findAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	struct tagCfg *av, fav;
	register int cx = 0;

	if (!cfg) {
		cfg_SetErr(EINVAL, "Invalid argument(s)");
		return -1;
	} else
		memset(&fav, 0, sizeof fav);

	if (csSec && *csSec)
		AIT_KEY(&fav.cfg_sec) = crcFletcher16((u_short*) csSec, 
				E_ALIGN(strlen(csSec), 2) / 2);
	if (csAttr)
		AIT_KEY(&fav.cfg_attr) = crcFletcher16((u_short*) csAttr, 
				E_ALIGN(strlen(csAttr), 2) / 2);

	TAILQ_FOREACH(av, cfg, cfg_next) {
		++cx;
		if (!cfg_tree_cmp(&fav, av))
			return cx;
	}

	return 0;
}

/*
 * cfg_unsetAttribute() - Unset item from config and free resources
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: 0 item not found, -1 error or 1 removed item
 */
int
cfg_unsetAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	struct tagCfg *av;

	if (!cfg)
		return -1;

	av = _selectAttribute(cfg, csSec, csAttr);
	if (!av)
		return 0;

	CFG_RC_LOCK(cfg);
	RB_REMOVE(tagRC, cfg, av);
	TAILQ_REMOVE(cfg, av, cfg_next);
	CFG_RC_UNLOCK(cfg);

	AIT_FREE_VAL(&av->cfg_val);
	AIT_FREE_VAL(&av->cfg_attr);
	AIT_FREE_VAL(&av->cfg_sec);
	e_free(av);
	return 1;
}

/*
 * cfg_setAttribute() - Set item in config or adding new item if not exists
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL set in *default* section
 * @csAttr = Config attribute //{csAttr} = ...
 * @csVal = Config value //... = {csVal} to setup
 * return: 0 nothing changed, -1 error, 1 found and updated item or 2 added new item
 */
int
cfg_setAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr, const char *csVal)
{
	struct tagCfg *av, *section;

	if (!cfg || !csAttr)
		return -1;

	av = _selectAttribute(cfg, csSec, csAttr);
	if (!av) {
		/* adding new element */
		section = _selectAttribute(cfg, csSec, NULL);

		av = e_malloc(sizeof(struct tagCfg));
		if (!av) {
			LOGERR;
			return -1;
		} else {
			memset(av, 0, sizeof(struct tagCfg));

			CFG_RC_LOCK(cfg);
			if (!section)
				TAILQ_INSERT_TAIL(cfg, av, cfg_next);
			else
				TAILQ_INSERT_BEFORE(section, av, cfg_next);
			CFG_RC_UNLOCK(cfg);
		}

		if (csSec && *csSec) {
			AIT_SET_STR(&av->cfg_sec, csSec);
			AIT_KEY(&av->cfg_sec) = crcFletcher16(AIT_GET_LIKE(&av->cfg_sec, u_short*), 
					E_ALIGN(AIT_LEN(&av->cfg_sec) - 1, 2) / 2);
		}
		AIT_SET_STR(&av->cfg_val, csVal ? csVal : "");
		AIT_SET_STR(&av->cfg_attr, csAttr);
		AIT_KEY(&av->cfg_attr) = crcFletcher16(AIT_GET_LIKE(&av->cfg_attr, u_short*), 
				E_ALIGN(AIT_LEN(&av->cfg_attr) - 1, 2) / 2);

		CFG_RC_LOCK(cfg);
		RB_INSERT(tagRC, cfg, av);
		CFG_RC_UNLOCK(cfg);
		return 2;
	}

	if (csVal && AIT_ADDR(&av->cfg_val) && 
			strcmp((char*) csVal, (char*) AIT_GET_STR(&av->cfg_val))) {
		/* Update element */
		AIT_FREE_VAL(&av->cfg_val);
		AIT_SET_STR(&av->cfg_val, csVal);
		return 1;
	}

	/* Nothing happens ... found & values is equal! */
	return 0;
}

/*
 * cfg_getAttribute() - Get item from config and return value from it
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: NULL item not found or null parameters, !=NULL value const string
 */
const char *
cfg_getAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	struct tagCfg *av;

	if (!cfg)
		return NULL;

	av = _selectAttribute(cfg, csSec, csAttr);
	if (!av)
		return NULL;

	return AIT_GET_STR(&av->cfg_val);
}

/*
 * cfg_getAttributeLong() - Get item as long from config and return value from it
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: value
 */
long
cfg_getAttributeLong(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	const char *str = NULL;

	str = cfg_getAttribute(cfg, csSec, csAttr);
	return strtol(str ? str : "", NULL, 0);
}

/*
 * cfg_getAttributeLLong() - Get item as long long from config and return value from it
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: value
 */
long long
cfg_getAttributeLLong(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	const char *str = NULL;

	str = cfg_getAttribute(cfg, csSec, csAttr);
	return strtoll(str ? str : "", NULL, 0);
}

/*
 * cfg_getAttributeDouble() - Get item as double from config and return value from it
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: value
 */
double
cfg_getAttributeDouble(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	const char *str = NULL;

	str = cfg_getAttribute(cfg, csSec, csAttr);
	return strtod(str ? str : "", NULL);
}

/*
 * cfg_getAttributeLDouble() - Get item as long double from config and return value from it
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * return: value
 */
long double
cfg_getAttributeLDouble(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr)
{
	const char *str = NULL;

	str = cfg_getAttribute(cfg, csSec, csAttr);
	return strtold(str ? str : "", NULL);
}

/*
 * cfg_loadAttribute() - Get guarded attribute, if not found item return *default value*
 *
 * @cfg = Config root
 * @csSec = Config section //[{csSec}], if NULL unset in *default* section
 * @csAttr = Config attribute //{csAttr} = ..., if NULL as *any* attribute
 * @val = Return buffer for item Value //... = {val}
 * @csDefValue = *Default Value* for return in //{val}, if not found item in config
 * return: 0 item not found, -1 error or >0 number of copied bytes in //{val}
 */
int
cfg_loadAttribute(cfg_root_t * __restrict cfg, const char *csSec, const char *csAttr, 
		ait_val_t * __restrict val, const char *csDefValue)
{
	struct tagCfg *av;
	int ret = 0;

	if (!cfg || !val) {
		cfg_SetErr(EINVAL, "Invalid argument(s)");
		return -1;
	}

	AIT_INIT_VAL(val);
	av = _selectAttribute(cfg, csSec, csAttr);
	if (!av) {
		/* not found item */
		if (csDefValue) {
			AIT_SET_STR(val, csDefValue);
			ret = AIT_LEN(val);
		} else
			AIT_INIT_VAL(val);
		return ret;
	}

	if (AIT_ISEMPTY(&av->cfg_val) || !AIT_ADDR(&av->cfg_val) || 
			!*AIT_GET_LIKE(&av->cfg_val, char*)) {
		/* empty value */
		if (csDefValue) {
			AIT_SET_STR(val, csDefValue);
			ret = AIT_LEN(val);
		} else
			AIT_INIT_VAL(val);
	} else {
		/* copy value */
		AIT_SET_STR(val, AIT_GET_STR(&av->cfg_val));
		ret = AIT_LEN(val);
	}

	return ret;
}
